
/*******************************************************************************
*
* File star1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the quadratic part of the gauge-fixing action density.
*
*   void star1(su3_alg_dble *ad,double *star)
*     Computes the leading-order contribution to the gauge-fixing action
*     density as a function of the time x0 (see the notes). The result is
*     averaged over the space coordinates.
*
* For both SF and open-SF boundary conditions and all times 0<x0<N0, the
* values computed by star1() are
*
*  star[x0]=(N1*N2*N3)^(-1)*sum_{x1,x2,x3}{sum_mu[A^a_mu(x)-A^a_mu(x-mu)]}^2,
*
* where N0=NPROC0*L0,..,N3=NPROC3*L3 and A_mu(x) is the gauge potential
* defined by the field array ad.
*
* For open-SF boundary conditions the value computed at x0=0 is
*
*  star[0]=(2*N1*N2*N3)^(-1)*sum_{x1,x2,x3}
*
*          *{2*A^a_0(x)+sum_{mu>0}[A^a_mu(x)-A^a_mu(x-mu)]}^2,
*
* while for SF boundary conditions
*
*  star[0]=(N0^3*N1*N2*N3)^(-1)*sum_{y,z}A^a_0(y)A^a_0(z).
*
* The program star1() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define STAR1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static double loc_star[N0];
static const su3_alg_dble ad0={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
static su3_alg_dble A0,sA0;


static void set_A0(su3_alg_dble *ad)
{
   int ix;

   A0=ad0;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      _su3_alg_add_assign(A0,*ad);
      ad+=1;
      _su3_alg_add_assign(A0,*ad);
      ad+=7;
   }
}


static void local_star(su3_alg_dble *cd)
{
   int ix,t;

   for (t=0;t<N0;t++)
      loc_star[t]=0.0;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);
      loc_star[t]+=su3alg_biform(cd,cd);
      cd+=1;
   }
}


void star1(su3_alg_dble *ad,double *star)
{
   int bc,t;
   double r0;
   su3_alg_dble **wcd;

   wcd=reserve_wcd(1);
   divad(ad,wcd[0]);
   local_star(wcd[0]);
   MPI_Reduce(loc_star,star,N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(star,N0,MPI_DOUBLE,0,MPI_COMM_WORLD);

   r0=1.0/((double)(N1*N2*N3));

   for (t=1;t<N0;t++)
      star[t]*=r0;

   bc=bc_type();

   if (bc==1)
   {
      set_A0(ad);
      MPI_Reduce(&A0,&sA0,8,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sA0,8,MPI_DOUBLE,0,MPI_COMM_WORLD);
      r0/=(double)(N0*N0*N0);
      star[0]=r0*su3alg_biform(&sA0,&sA0);
   }
   else
      star[0]*=(0.5*r0);

   release_wcd();
}
