
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2005, 2007, 2010, 2011, 2013, 2015 Martin Luescher,
*                                                  Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Writing and reading of gauge field configurations.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "linalg.h"
#include "uflds.h"
#include "lattice.h"
#include "su3fcts.h"
#include "archive.h"
#include "global.h"

#define NMAX 4

typedef union
{
   su3_dble u;
   double r[18];
} umat_t;

static int *rlxs_state[2],*rlxd_state[2];


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));

   mulr_alg(4*VOLUME,0.1,ad,ad);
   set_adbc(0,ad);
}


static void random_ufld(int n,su3_dble *ud)
{
   int i;
   su3_alg_dble **wad;

   wad=reserve_wad(n+1);

   for (i=0;i<=n;i++)
      new_fld(wad[i]);

   a2uflds(n,wad,ud);
   release_wad();
}


static void save_ranlux(void)
{
   int nlxs,nlxd;
   int *p;

   nlxs=rlxs_size();
   nlxd=rlxd_size();

   p=malloc(2*(nlxs+nlxd)*sizeof(*p));
   error(p==NULL,1,"save_ranlux [check1.c]",
         "Unable to allocate state arrays");
   rlxs_state[0]=p;
   p+=nlxs;
   rlxs_state[1]=p;
   p+=nlxs;
   rlxd_state[0]=p;
   p+=nlxd;
   rlxd_state[1]=p;

   rlxs_get(rlxs_state[0]);
   rlxd_get(rlxd_state[0]);
}


static int check_ranlux(void)
{
   int nlxs,nlxd,k,ie;

   nlxs=rlxs_size();
   nlxd=rlxd_size();

   rlxs_get(rlxs_state[1]);
   rlxd_get(rlxd_state[1]);
   ie=0;

   for (k=0;k<nlxs;k++)
      ie|=(rlxs_state[0][k]!=rlxs_state[1][k]);

   for (k=0;k<nlxd;k++)
      ie|=(rlxd_state[0][k]!=rlxd_state[1][k]);

   return ie;
}


static int is_equal(double tol,su3_dble *ud1,su3_dble *ud2)
{
   int i,it;
   umat_t *um,*vm;

   um=(umat_t*)(ud1);
   vm=(umat_t*)(ud2);
   it=1;

   for (i=0;i<18;i++)
      it&=(fabs((*um).r[i]-(*vm).r[i])<=tol);

   return it;
}


static int cmp_ud(double tol,int n,su3_dble *ud1,su3_dble *ud2)
{
   int ie;
   su3_dble *udm;

   udm=ud1+4*VOLUME*(n+1);
   ie=1;

   for (;ud1<udm;ud1++)
   {
      ie&=is_equal(tol,ud1,ud2);
      ud2+=1;
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,ie;
   su3_dble **wud;
   char loc_dir[NAME_SIZE],cnfg[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check1.log",0);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
      fin=freopen("check1.in","r",stdin);

      printf("\n");
      printf("Writing and reading of gauge field configurations\n");
      printf("-------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("loc_dir","%s\n",loc_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check1.c]",
                    "Syntax: check1 [-bc <type>]");
   }

   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   check_dir(loc_dir);
   nsize=name_size("%s/testcnfg_%d",loc_dir,NPROC);
   error_root(nsize>=NAME_SIZE,1,"main [check1.c]","loc_dir name is too long");
   sprintf(cnfg,"%s/testcnfg_%d",loc_dir,my_rank);

   alloc_wud(2*NMAX);
   alloc_wad(1*NMAX);

   if (my_rank==0)
   {
      printf("Write random field configuration to the files\n"
             "%s/testcnfg_*\n"
             "on the local disks.\n\n",loc_dir);
      printf("Then read the field from there, compare with the saved field\n"
             "and remove all files.\n\n");
   }

   for (n=0;n<NMAX;n++)
   {
      wud=reserve_wud(2*(n+1));
      random_ufld(n,wud[0]);

      cm3x3_assign((4*VOLUME+7*(BNDRY/4))*(n+1),wud[0],wud[n+1]);
      save_ranlux();
      write_cnfg(cnfg,n,wud[0]);

      random_ufld(n,wud[0]);
      read_cnfg(cnfg,n,wud[0]);
      remove(cnfg);

      ie=check_udbc(n,wud[0]);
      ie&=cmp_ud(64.0*DBL_EPSILON,n,wud[0],wud[n+1]);
      error(ie!=1,1,"main [check1.c]","The link field is not properly restored");

      ie=check_ranlux();
      error(ie!=0,1,"main [check1.c]",
            "The random number generator is not properly restored");

      release_wud();
   }

   if (my_rank==0)
   {
      printf("No errors detected --- the fields are correctly written\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
