
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005, 2007, 2008, 2010, 2011, 2013,  Martin Luescher,
*               2015, 2016                           Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Importing gauge configurations previously exported by check3.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "linalg.h"
#include "uflds.h"
#include "lattice.h"
#include "su3fcts.h"
#include "archive.h"
#include "global.h"

#define NMAX 4

typedef union
{
   su3_dble u;
   double r[18];
} umat_t;


static void normalized_action(int n,su3_dble *ud,double *S)
{
   int k;
   double nplaq;

   nplaq=((double)(6*NPROC)*(double)(VOLUME));
   Splaq(1,n,ud,S);

   for (k=0;k<=n;k++)
      S[k]/=nplaq;
}


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));

   mulr_alg(4*VOLUME,0.1,ad,ad);
   set_adbc(0,ad);
}


static void random_ufld(int n,su3_dble *ud)
{
   int i;
   su3_alg_dble **wad;

   wad=reserve_wad(n+1);

   for (i=0;i<=n;i++)
      new_fld(wad[i]);

   a2uflds(n,wad,ud);
   release_wad();
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,nsize,k,ir,ie;
   stdint_t istd[5];
   double *act0,*act1,*act2,dev;
   su3_dble **wud;
   char cnfg_dir[NAME_SIZE],cnfg[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check3.log",0);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Importing gauge configurations exported by check3\n");
      printf("-------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("cnfg_dir","%s\n",cnfg_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
   }

   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,9876);
   geometry();

   check_dir_root(cnfg_dir);
   nsize=name_size("%s/testcnfgn%d",cnfg_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check3.c]","cnfg_dir name is too long");

   alloc_wud(NMAX);
   alloc_wad(NMAX);
   act0=malloc(3*NMAX*sizeof(*act0));
   act1=act0+NMAX;
   act2=act1+NMAX;

   for (n=0;n<NMAX;n++)
   {
      wud=reserve_wud(n+1);
      random_ufld(n,wud[0]);
      normalized_action(n,wud[0],act0);

      sprintf(cnfg,"%s/testcnfgn%d",cnfg_dir,n);

      if (my_rank==0)
      {
         fin=fopen(cnfg,"rb");
         error_root(fin==NULL,1,"main [check3.c]","Unable to open input file");

         ir=fread(istd,sizeof(stdint_t),5,fin);
         ir+=fread(act1,sizeof(double),n+1,fin);
         error_root(ir!=6+n,1,"main [check3.c]","Incorrect read count");
         fclose(fin);

         if (endianness()==BIG_ENDIAN)
         {
            bswap_int(5,istd);
            bswap_double(n+1,act1);
         }

         printf("Random gauge field of order %d, plaquette action:\n",n);
         for (k=0;k<=n;k++)
            printf("k = %2d, act = %+.15e\n",k,act0[k]);

         printf("\n");
         printf("Now read gauge field from file\n"
                "%s:\n",cnfg);
         printf("%dx%dx%dx%d lattice, num. of field-coefficients %d\n\n",
                (int)(istd[0]),(int)(istd[1]),(int)(istd[2]),
                (int)(istd[3]),(int)(istd[4]));
         printf("Plaquette action:\n");
         for (k=0;k<=n;k++)
            printf("k = %2d, act = %+.15e\n",k,act1[k]);

         printf("\n");
      }

      import_cnfg(cnfg,n,wud[0]);
      ie=check_udbc(n,wud[0]);
      error(ie!=1,1,"main [check3.c]","Boundary conditions are not preserved");
      normalized_action(n,wud[0],act2);

      if (my_rank==0)
      {
         printf("Should be:\n");
         for (k=0;k<=n;k++)
         {
            dev=fabs(1.0-act2[k]/act1[k]);
            printf("k = %2d, act = %+.15e, relative deviation = %.1e\n",
                   k,act2[k],dev);
         }

         printf("\n\n");
         remove(cnfg);
      }

      release_wud();
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
