
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015-2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Stability of the MD evolution.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "forces.h"
#include "mdint.h"
#include "global.h"

static int my_rank;
static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static su3_alg_dble *Aw=NULL;
static su3_dble *Uw;


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check3.c]",
         "Unable to allocate auxiliary array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void read_smd_parms(void)
{
   int nlv,nact,iact[1];
   double eps,gamma,lambda0;

   if (my_rank==0)
   {
      find_section("SMD parameters");
      read_line("nlv","%d",&nlv);
      read_line("eps","%lf",&eps);
      read_line("gamma","%lf",&gamma);
      read_line("lambda0","%lf",&lambda0);
   }

   MPI_Bcast(&nlv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&gamma,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   nact=1;
   iact[0]=0;
   set_smd_parms(nact,iact,0,0,NULL,nlv,eps,gamma,lambda0);
}


static void read_nspt_parms(void)
{
   int bc,order;
   int ncG,ncGP,i;
   double *cG,*cGP;

   bc=bc_type();

   if (my_rank==0)
   {
      find_section("NSPT parameters");
      read_line("order","%d",&order);
      ncG=count_tokens("cG");
      if (bc==2)
         ncGP=count_tokens("cG'");
      else
         ncGP=0;
   }

   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncG,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncGP,1,MPI_INT,0,MPI_COMM_WORLD);

   if (ncG>0)
   {
      cG=malloc(ncG*sizeof(*cG));
      error(cG==NULL,1,"read_nspt_parms [check3.c]",
                 "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG",ncG,cG);
      }

      MPI_Bcast(cG,ncG,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cG=NULL;

   if (ncGP>0)
   {
      cGP=malloc(ncGP*sizeof(*cGP));
      error(cGP==NULL,1,"read_nspt_parms [check3.c]",
                 "Unable to allocate auxiliary array");

      if (my_rank==0)
      {
         find_section("NSPT parameters");
         read_dprms("cG'",ncGP,cGP);
      }

      MPI_Bcast(cGP,ncGP,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      cGP=NULL;

   default_pv(order);

   for (i=2;(i<=(2*ncG))&&(i<=order);i+=2)
   {
      opv[(int)(CG)]=i;
      pv[(int)(CG)][i]=cG[i/2-1];
   }

   for (i=2;(i<=(2*ncGP))&&(i<=order);i+=2)
   {
      opv[(int)(CGP)]=i;
      pv[(int)(CGP)][i]=cGP[i/2-1];
   }

   set_nspt_parms(KSPT,order,opv,pv);
}


static void read_integrator(void)
{
   int nlv,i;
   smd_parms_t smd;

   smd=smd_parms();
   nlv=smd.nlv;

   for (i=0;i<nlv;i++)
      read_mdint_parms(i);
}


static void read_run_parms(int *nth,int *ntr)
{
   if (my_rank==0)
   {
      find_section("Run parameters");
      read_line("nth","%d",nth);
      read_line("ntr","%d",ntr);
   }

   MPI_Bcast(nth,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(ntr,1,MPI_INT,0,MPI_COMM_WORLD);
}


static void dev_lnud(int n,su3_dble *ud,double (*dev)[2])
{
   int i;
   double d;

   cm3x3_log_poly(n+1,ud,Uw);

   for (i=0;i<=n;i++)
   {
      su3_to_su3alg(Uw+i,Aw);
      d=sqrt(su3alg_biform(Aw,Aw));

      if (d>dev[i][0])
         dev[i][0]=d;

      dev[i][1]+=d;
   }
}


static void dev_ud(int n,su3_dble *ud,double (*dev)[2])
{
   int m,ix,ifc,i;
   double d[2];

   m=n+1;

   if (Aw==NULL)
   {
      Aw=amalloc(1*sizeof(*Aw),ALIGN);
      Uw=amalloc(m*sizeof(*Uw),ALIGN);
      error((Aw==NULL)||(Uw==NULL),1,"dev_ud [check3.c]",
            "Unable to allocate auxiliary arrays");
   }

   for (i=0;i<=n;i++)
   {
      dev[i][0]=0.0;
      dev[i][1]=0.0;
   }

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         dev_lnud(n,ud,dev);
         ud+=m;
      }
   }

   if (NPROC>1)
   {
      for (i=0;i<=n;i++)
      {
         d[0]=dev[i][0];
         d[1]=dev[i][1];
         MPI_Reduce(d,dev[i],1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
         MPI_Reduce(d+1,dev[i]+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(dev[i],2,MPI_DOUBLE,0,MPI_COMM_WORLD);
      }
   }

   for (i=0;i<=n;i++)
      dev[i][1]/=((double)(4*NPROC)*(double)(VOLUME));
}


static void dev_su3_ud(int n,su3_dble *ud,double *dev)
{
   int m,ix,ifc;
   double d;

   m=n+1;
   dev[0]=0.0;
   dev[1]=0.0;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         d=cm3x3_devsu3_poly(m,ud);
         if (d>dev[0])
            dev[0]=d;
         dev[1]+=d;
         ud+=m;
      }
   }

   if (NPROC>1)
   {
      d=dev[0];
      MPI_Reduce(&d,dev,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(dev,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      d=dev[1];
      MPI_Reduce(&d,dev+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(dev+1,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   dev[1]/=((double)(4*NPROC)*(double)(VOLUME));
}


int main(int argc,char *argv[])
{
   int bc,n,nth,ntr;
   int i,j;
   double *dH,(*mA)[2],dev[3];
   su3_alg_dble **wad,*mom;
   su3_alg_dble **wcd,*om;
   su3_dble **wud,*ud;
   nspt_parms_t nspt;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);

      printf("\n");
      printf("Stability of the MD evolution\n");
      printf("-----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);

   read_smd_parms();
   read_nspt_parms();
   read_integrator();
   read_run_parms(&nth,&ntr);

   if (my_rank==0)
      fclose(fin);

   print_th_parms();
   print_nspt_parms((1<<(int)(LAMBDA))^0xffff);
   print_smd_parms();
   print_mdint_parms();

   start_ranlux(0,12345);
   geometry();

   nspt=nspt_parms();
   n=nspt.order;

   alloc_wad(2*(n+1));
   alloc_wud(n+1);
   alloc_wcd(2*(n+1));
   alloc_wgd(n+1);

   wad=reserve_wad(n+1);
   wud=reserve_wud(n+1);
   wcd=reserve_wcd(n+1);

   mom=wad[0];
   ud=wud[0];
   om=wcd[0];

   dH=malloc((n+1)*sizeof(*dH));
   mA=malloc((n+1)*sizeof(*mA));
   error((dH==NULL)||(mA==NULL),1,"main [check3.c]",
         "Unable to allocate auxiliary arrays");

   set_mdsteps();
   print_mdsteps(0x3);
   start_smd(n,om,mom,ud);

   if (my_rank==0)
      fflush(flog);

   for (i=0;i<nth;i++)
      run_mdint(n,om,mom,ud,0,NULL);

   for (i=0;i<ntr;i++)
   {
      run_mdint(n,om,mom,ud,1,dH);
      dev_su3_ud(n,ud,dev);
      dev_ud(n,ud,mA);

      if (my_rank==0)
      {
         printf("Trajectory no %d\n",i+1);
         printf("max|ud-su3| = %.1e, avg|ud-su3| = %.1e\n",dev[0],dev[1]);

         for (j=0;j<=n;j++)
            printf("order = %d, max|A| = %.1e, avg|A| = %.1e, dH = % .2e\n",
                   j,mA[j][0],mA[j][1],dH[j]);

         printf("\n");
         fflush(flog);
      }
   }

   dev[2]=renormalize_ud(n,ud);
   dev_su3_ud(n,ud,dev);

   if (my_rank==0)
   {
      printf("After renormalization to SU(3)\n");
      printf("max|ud-su3| = %.1e, avg|ud-su3| = %.1e, ",dev[0],dev[1]);
      printf("max|ud-ud_old| = %.1e\n\n",dev[2]);
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
