/*******************************************************************************
*
* File cdcom.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Communication of the ghost field variables at the exterior boundaries of
* the local lattices.
*
*   void cpcd_int_bnd(int n,su3_alg_dble *cd)
*     Copies the field cd on the even points along the interior boundaries
*     of the local lattice to the corresponding exterior boundary points
*     on the neighbouring MPI processes. The meaning of the parameter n>=0
*     is explained in the notes.
*
*   void cpcd_ext_bnd(int n,su3_alg_dble *cd)
*     Adds the field cd at the even exterior boundary points to the field
*     on the matching lattice points on the neighbouring MPI processes. The
*     meaning of the parameter n>=0 is explained in the notes.
*
* In these programs, the field array cd is assumed to contain n+1 ghost
* fields of the kind returned by reserve_wcd(n+1) (see utils/wspace.c). More
* precisely, the fields are expected to be arranged in memory such that the
* field variables at the point number i are cd[i*(n+1)+k], where k=0,..,n is
* the field index.
*
* NOTE: Across the boundary of the lattice in the time direction, the field
* variables are copied as if periodic boundary conditions were imposed. It is
* up to the calling program to ensure that the chosen boundary conditions are
* respected.
*
* All these programs involve global communications and must be called on all
* processes simultaneously.
*
*******************************************************************************/

#define CDCOM_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"

static int ms=0,np,nfc[8],ofs[8],tag0[8];
static su3_alg_dble *buf,*sbuf,*rbuf;


static void alloc_buf(int n)
{
   int m,nbf,ifc;

   if (ms==0)
   {
      np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

      nfc[0]=FACE0/2;
      nfc[1]=FACE0/2;
      nfc[2]=FACE1/2;
      nfc[3]=FACE1/2;
      nfc[4]=FACE2/2;
      nfc[5]=FACE2/2;
      nfc[6]=FACE3/2;
      nfc[7]=FACE3/2;

      ofs[0]=0;
      ofs[1]=ofs[0]+FACE0/2;
      ofs[2]=ofs[1]+FACE0/2;
      ofs[3]=ofs[2]+FACE1/2;
      ofs[4]=ofs[3]+FACE1/2;
      ofs[5]=ofs[4]+FACE2/2;
      ofs[6]=ofs[5]+FACE2/2;
      ofs[7]=ofs[6]+FACE3/2;

      for (ifc=0;ifc<8;ifc++)
         tag0[ifc]=mpi_permanent_tag();
   }

   m=n+1;
   nbf=0;

   for (ifc=0;ifc<8;ifc++)
   {
      if (nbf<nfc[ifc])
         nbf=nfc[ifc];
   }

   if (ms>0)
      afree(buf);

   buf=amalloc(nbf*m*sizeof(*buf),ALIGN);
   error(buf==NULL,1,"alloc_buf [cdcom.c]",
         "Unable to allocate communication buffer");

   ms=m;
}


static void pack_bnd_cd(int n,int ifc,su3_alg_dble *cd)
{
   int m,*ic,*icm;
   su3_alg_dble *X;

   m=n+1;
   ic=map+ofs[ifc^0x1];
   icm=ic+nfc[ifc];
   X=buf;

   for (;ic<icm;ic++)
   {
      assign_alg2alg(m,cd+ic[0]*m,X);
      X+=m;
   }
}


static void unpack_bnd_cd(int n,int ifc,su3_alg_dble *cd)
{
   int m,*ic,*icm;
   su3_alg_dble *X;

   m=n+1;
   ic=map+ofs[ifc];
   icm=ic+nfc[ifc];
   X=buf;

   for (;ic<icm;ic++)
   {
      add_alg(m,X,cd+ic[0]*m);
      X+=m;
   }
}


static void send_bnd_cd(int n,int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag0[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=8*nfc[ifc]*(n+1);

   if (np==0)
   {
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


void cpcd_int_bnd(int n,su3_alg_dble *cd)
{
   int m,ifc,ncd;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=buf;
      rbuf=cd+VOLUME*m;

      for (ifc=0;ifc<8;ifc++)
      {
         ncd=nfc[ifc];

         if (ncd>0)
         {
            pack_bnd_cd(n,ifc^0x1,cd);
            send_bnd_cd(n,ifc^0x1);
            rbuf+=ncd*m;
         }
      }
   }
}


void cpcd_ext_bnd(int n,su3_alg_dble *cd)
{
   int m,ifc,ncd;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=cd+VOLUME*m;
      rbuf=buf;

      for (ifc=0;ifc<8;ifc++)
      {
         ncd=nfc[ifc];

         if (ncd>0)
         {
            send_bnd_cd(n,ifc);
            unpack_bnd_cd(n,ifc,cd);
            sbuf+=ncd*m;
         }
      }
   }
}
