
/*******************************************************************************
*
* File cdshift.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Translation of the ghost fields.
*
*   int shift_cd(int *s,su3_alg_dble *cd)
*     Replaces the ghost field C(x), represented by the array cd, by C(x-s),
*     where s[4] is a given shift vector. The program returns the number of
*     elementary steps (translations by 1 lattice unit) that were performed.
*
* Shifts in the time direction are not permitted. The array cd is assumed to
* be of size at least NSPIN like the fields returned by the workspace programs
* (utils/wspace.c).
*
* The program shift_cd() acts globally and must be called simultaneously on
* all MPI processes with the same translation vector.
*
*******************************************************************************/

#define CDSHIFT_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "utils.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

static int init=0,bs[4],np[4],npt[4];
static int **idx[8];
static int isnd,tag0[8];
static su3_alg_dble *sbuf,*rbuf;


static void set_const(void)
{
   bs[0]=L0;
   bs[1]=L1;
   bs[2]=L2;
   bs[3]=L3;

   np[0]=NPROC0;
   np[1]=NPROC1;
   np[2]=NPROC2;
   np[3]=NPROC3;

   npt[0]=VOLUME/L0;
   npt[1]=VOLUME/L1;
   npt[2]=VOLUME/L2;
   npt[3]=VOLUME/L3;
}


static void alloc_idx(void)
{
   int ifc,mu,t;
   int **idd,*id;

   idd=malloc(2*(bs[0]+bs[1]+bs[2]+bs[3])*sizeof(*idd));
   id=malloc(4*VOLUME*sizeof(*id));
   error((idd==NULL)||(id==NULL),1,"alloc_idx [cdshift.c]",
         "Unable to allocate index arrays");

   for (ifc=0;ifc<8;ifc++)
   {
      mu=ifc/2;
      idx[ifc]=idd;
      idd+=bs[mu];

      if ((ifc&0x1)==0)
      {
         for (t=0;t<bs[mu];t++)
         {
            idx[ifc][t]=id;
            id+=npt[mu];
         }
      }
      else
      {
         for (t=0;t<bs[mu];t++)
            idx[ifc][t]=idx[ifc-1][bs[mu]-1-t];
      }
   }
}


static void set_idx(void)
{
   int x0,x1,x2,x3,x[4];
   int mu,ix,iy[4];

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               x[0]=x0;
               x[1]=x1;
               x[2]=x2;
               x[3]=x3;

               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               iy[0]=x3+L3*x2+L2*L3*x1;
               iy[1]=x0+L0*x3+L3*L0*x2;
               iy[2]=x1+L1*x0+L0*L1*x3;
               iy[3]=x2+L2*x1+L1*L2*x0;

               for (mu=0;mu<4;mu++)
                  idx[2*mu][x[mu]][iy[mu]]=ix;
            }
         }
      }
   }
}


static void alloc_buf(void)
{
   int ifc,ncd;

   isnd=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;
   ncd=0;

   for (ifc=0;ifc<8;ifc++)
   {
      tag0[ifc]=mpi_permanent_tag();

      if (npt[ifc/2]>ncd)
         ncd=npt[ifc/2];
   }

   sbuf=amalloc(ncd*sizeof(*sbuf),ALIGN);
   error(sbuf==NULL,1,"alloc_buf [cdshift.c]",
         "Unable to allocate communication buffer");
}


static void pack_bnd_cd(int ifc,su3_alg_dble *cd)
{
   int *ic,*icm;
   su3_alg_dble *X;

   ic=idx[ifc][0];
   icm=ic+npt[ifc/2];
   X=sbuf;

   for (;ic<icm;ic++)
   {
      (*X)=cd[*ic];
      X+=1;
   }
}


static void unpack_bnd_cd(int ifc,su3_alg_dble *cd)
{
   int *ic,*icm;
   su3_alg_dble *X;

   ic=idx[ifc][bs[ifc/2]-1];
   icm=ic+npt[ifc/2];
   X=rbuf;

   for (;ic<icm;ic++)
   {
      cd[*ic]=(*X);
      X+=1;
   }
}


static void send_bnd_cd(int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag0[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=8*npt[ifc/2];

   if (isnd==0)
   {
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


static void shift_step(int ifc,su3_alg_dble *cd)
{
   int t,*id0,*id1,*idm;

   pack_bnd_cd(ifc,cd);

   for (t=0;t<(bs[ifc/2]-1);t++)
   {
      id0=idx[ifc][t];
      id1=idx[ifc][t+1];
      idm=id0+npt[ifc/2];

      for (;id0<idm;id0++,id1++)
         cd[*id0]=cd[*id1];
   }

   if (np[ifc/2]==1)
      rbuf=sbuf;
   else
   {
      rbuf=cd+VOLUME;
      send_bnd_cd(ifc);
   }

   unpack_bnd_cd(ifc,cd);
}


int shift_cd(int *s,su3_alg_dble *cd)
{
   int iprms[4],sr[4];
   int mu,ifc,t,n;

   if (NPROC>1)
   {
      iprms[0]=s[0];
      iprms[1]=s[1];
      iprms[2]=s[2];
      iprms[3]=s[3];

      MPI_Bcast(iprms,4,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=s[0])||(iprms[1]!=s[1])||
            (iprms[2]!=s[2])||(iprms[3]!=s[3]),1,
            "shift_ad [cdshift.c]","Shift vector is not global");
   }

   if (init==0)
   {
      set_const();
      alloc_idx();
      set_idx();
      alloc_buf();
      init=1;
   }

   for (mu=0;mu<4;mu++)
   {
      n=np[mu]*bs[mu];

      if (abs(s[mu])>(n/2))
      {
         sr[mu]=safe_mod(s[mu],n);

         if (sr[mu]>(n/2))
            sr[mu]-=n;
      }
      else
         sr[mu]=s[mu];
   }

   if ((sr[0]==0)&&(sr[1]==0)&&(sr[2]==0)&&(sr[3]==0))
      return 0;

   error_root(sr[0]!=0,1,"shift_cd [cdshift.c]",
              "Shifts in time are not permitted for SF and open-SF bc");

   n=0;

   for (mu=0;mu<4;mu++)
   {
      if (sr[mu]>=0)
         ifc=2*mu+1;
      else
         ifc=2*mu;

      for (t=0;t<abs(sr[mu]);t++)
      {
         shift_step(ifc,cd);
         n+=1;
      }
   }

   return n;
}
