
/*******************************************************************************
*
* File force0.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Gauge action and associated MD force.
*
* The externally accessible functions are
*
*   void force0(int n,su3_dble *ud,double c,su3_alg_dble *mom)
*     Computes the force deriving from the gauge action of the link field
*     ud and adds the force multiplied by c to the momentum field mom.
*
*   void action0(int icom,int n,su3_dble *ud,double *S)
*     Assigns the gauge action of the link field ud to S (see the notes).
*     The local parts of the action are summed over all MPI processes if
*     (and only if) icom=1.
*
* See the notes doc/gauge_action.pdf for the definition of the gauge action.
* Currently only the Wilson action (including the boundary improvement terms
* proportional to cG and cG') is supported. The expansion of cG and cG' in
* powers of g0 is retrieved from the parameter data base [pmdb/nspt_parms.c].
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* array fd similarly contains the n+1 coefficient fields in the expansion
*
*   F=g0*F[0]+g0^2*F[1]+..+g0^(n+1)*F[n]
*
* of the force field.
*
* The normalization and sign of the force field is such that
*
*   F(x,mu)=w(x,mu)*{dS(t)/dt}_{t=0}*T^a,
*
*   S(t)=S|_{U(x,mu)->exp(t*T^a)*U(x,mu)},
*
* where
*
*   w(x,mu) = 0  if x0=0, mu>0 and SF bc,
*
*           = 2  if x0=0, mu>0 and openSF bc,
*
*           = 1  otherwise.
*
* The values S[k], k=0,..,n, computed by action0() are the coefficients of
* g0^k in the perturbation expansion of the gauge action.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define FORCE0_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "su3fcts.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "forces.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ncG,ncGP,ms=0;
static int nsm[NMAX],ism[NMAX];
static double *cG,*cGP;
static su3_alg_dble *Aw;
static su3_dble *Vw,*Ww,*Xw,*Yw,*Uw[4];


static void set_cG(void)
{
   nspt_parms_t nsp;

   nsp=nspt_parms();

   if (nsp.opv[(int)(CG)]>0)
   {
      ncG=nsp.opv[(int)(CG)];
      cG=nsp.pv[(int)(CG)];
   }
   else
   {
      ncG=0;
      cG=NULL;
   }

   if (nsp.opv[(int)(CGP)]>0)
   {
      ncGP=nsp.opv[(int)(CGP)];
      cGP=nsp.pv[(int)(CGP)];
   }
   else
   {
      ncGP=0;
      cGP=NULL;
   }
}


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      set_cG();

      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }
   }

   if (ms>0)
   {
      afree(Aw);
      afree(Vw);
   }

   m=n+1;
   Aw=amalloc(m*sizeof(*Aw),ALIGN);
   Vw=amalloc(8*m*sizeof(*Vw),ALIGN);
   error((Aw==NULL)||(Vw==NULL),1,"alloc_arrays [force0.c]",
         "Unable to allocate auxiliary arrays");
   set_alg2zero(m,Aw);
   cm3x3_zero(8*m,Vw);

   Ww=Vw+m;
   Xw=Ww+m;
   Yw=Xw+m;
   Uw[0]=Yw+m;

   for (i=1;i<4;i++)
      Uw[i]=Uw[i-1]+m;

   ms=m;
}


static void mul_bnd_frc(int n,int nc,double *c)
{
   int i,j;

   for (i=n;i>=2;i--)
   {
      for (j=2;(j<=nc)&&(j<=i);j+=2)
         su3alg_mulr_add(c+j,Aw+i-j,Aw+i);
   }
}


void force0(int n,su3_dble *ud,double c,su3_alg_dble *mom)
{
   int bc,m,ix,t,i;
   int ncg,pln,ip[4];
   double *cg;

   bc=bc_type();
   m=n+1;
   ncg=0;
   cg=NULL;

   if (m>ms)
      alloc_arrays(n);

   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
   {
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);
      set_alg2zero((7*(BNDRY/4)+3)*m,mom+4*VOLUME*m);
   }
   else
      set_alg2zero(7*(BNDRY/4)*m,mom+4*VOLUME*m);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         if ((t!=0)||(pln<3)||(bc!=1))
         {
            ncg=0;

            if ((t==0)&&((bc==1)||(pln>=3)))
            {
               ncg=ncG;
               cg=cG;
            }
            else if ((t==(N0-1))&&(pln<3))
            {
               if (bc==1)
               {
                  ncg=ncG;
                  cg=cG;
               }
               else
               {
                  ncg=ncGP;
                  cg=cGP;
               }
            }

            plaq_aidx(pln,ix,ip);

            cm3x3_assign(m,ud+ip[0]*m,Uw[0]);
            cm3x3_assign(m,ud+ip[1]*m,Uw[1]);
            cm3x3_dagger(m,ud+ip[2]*m,Uw[2]);
            cm3x3_dagger(m,ud+ip[3]*m,Uw[3]);

            cm3x3_varmul_poly(m,Uw[3],Uw[2],Vw);
            cm3x3_varmul_poly(m,Uw[1],Vw,Ww);
            cm3x3_varmul_poly(m,Uw[0],Ww,Vw);

            for (i=0;i<m;i++)
               su3_to_su3alg(Vw+i,Aw+i);

            if (ncg)
               mul_bnd_frc(n,ncg,cg);
            mulr_add_alg(m,c,Aw,mom+ip[0]*m);

            if ((t!=0)||(pln>=3))
               mulr_add_alg(m,-c,Aw,mom+ip[2]*m);
            else if (bc!=1)
               mulr_add_alg(m,-2.0*c,Aw,mom+ip[2]*m);

            cm3x3_varmul_poly(m,Uw[2],Uw[0],Vw);
            cm3x3_varmul_poly(m,Uw[1],Uw[3],Ww);

            if ((t<(N0-1))||(pln>=3))
            {
               cm3x3_varmul_poly(m,Ww,Vw,Xw);

               for (i=0;i<m;i++)
                  su3_to_su3alg(Xw+i,Aw+i);

               if (ncg)
                  mul_bnd_frc(n,ncg,cg);
               mulr_add_alg(m,c,Aw,mom+ip[1]*m);
            }

            cm3x3_varmul_poly(m,Vw,Ww,Yw);

            for (i=0;i<m;i++)
               su3_to_su3alg(Yw+i,Aw+i);

            if (ncg)
               mul_bnd_frc(n,ncg,cg);
            mulr_add_alg(m,-c,Aw,mom+ip[3]*m);
         }
      }
   }

   add_bnd_ad(n,mom);
}


static int set_ism(int n)
{
   int m,i;

   m=n+1;

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [force0.c]","Requested too many hsum instances");

   return 0;
}


static double top_order(int n,su3_dble *X,su3_dble *Y)
{
   int j;
   double sm,tr;

   sm=0.0;

   for (j=0;j<=n;j++)
   {
      cm3x3_retr(X+j,Y+n-j,&tr);
      sm-=tr;
   }

   return 0.5*sm;
}


static void mul_bnd_act(int n,int nc,double *c,double *S)
{
   int i,j;

   for (i=n;i>=2;i--)
   {
      for (j=2;(j<=nc)&&(j<=i);j+=2)
         S[i]+=c[j]*S[i-j];
   }
}


void action0(int icom,int n,su3_dble *ud,double *S)
{
   int bc,m,ix,t,i;
   int is,pln,ip[4];

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);

   is=set_ism(n);
   reset_hsum(is);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         if ((pln<3)||(t>0)||(bc!=1))
         {
            plaq_aidx(pln,ix,ip);

            cm3x3_assign(m,ud+ip[0]*m,Uw[0]);
            cm3x3_assign(m,ud+ip[1]*m,Uw[1]);
            cm3x3_dagger(m,ud+ip[2]*m,Uw[2]);
            cm3x3_dagger(m,ud+ip[3]*m,Uw[3]);

            cm3x3_varmul_poly(m,Uw[3],Uw[2],Xw);
            cm3x3_varmul_poly(m,Uw[1],Xw,Yw);
            cm3x3_varmul_poly(m,Uw[0],Yw,Xw);

            for (i=1;i<m;i++)
               S[i-1]=Xw[i].c11.re+Xw[i].c22.re+Xw[i].c33.re;

            cm3x3_dagger(m,Xw,Yw);
            S[n]=top_order(n,Xw,Yw);

            if ((bc==1)&&((t==0)||(t==(N0-1)))&&(pln<3))
               mul_bnd_act(n,ncG,cG,S);
            else if ((bc==2)&&(t==0)&&(pln>=3))
            {
               mul_bnd_act(n,ncG,cG,S);

               for (i=0;i<m;i++)
                  S[i]*=0.5;
            }
            else if ((bc==2)&&(t==(N0-1))&&(pln<3))
               mul_bnd_act(n,ncGP,cGP,S);

            add_to_hsum(is,S);
         }
      }
   }

   if ((icom==1)&&(NPROC>1))
      global_hsum(is,S);
   else
      local_hsum(is,S);

   for (i=0;i<m;i++)
      S[i]=-2.0*S[i];
}
