
/*******************************************************************************
*
* File langevin.c
*
* Copyright (C) 2015, 2017 Mattia Dalla Brida, Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Integration of the Langevin equation.
*
*   void run_langevin(int n,su3_dble *ud)
*     Integrates the Langevin equations using the current integration
*     scheme (see the notes).
*
* The implementation of the Langevin equations is described in
*
*   M. Dalla Brida, "Numerical stochastic perturbation theory and the
*                    Langevin equations", notes (January 2016)
*
* [doc/langevin.pdf].
*
* In the program in this module, the array ud is expected to contain the n+1
* coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c).
*
* The Langevin equation is integrated according to the scheme specified
* in the Langevin parameter structure (see pmdb/lang_parms.c). Currently,
* an Euler scheme and a 2nd order Runge-Kutta are implemented.
*
* The workspace required depends on the choice of integrator. Specifically:
*
*  EULER:         n+1 su3_alg_dble link fields,
*
*  RK2:           n+2 su3_alg_dble link fields,
*                 n+1 su3_dble link fields,
*
* and additionally
*
*                 n+1 su3_alg_dble point fields,
*                 n+1 su3_dble point fields,
*
* for the gauge damping (see utils/wspace.c).
*
* The program in this module performs global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define LANGEVIN_C

#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "su3fcts.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "langevin.h"
#include "forces.h"
#include "mdint.h"
#include "global.h"

static int ms=0;
static su3_dble *Vw;
static su3_alg_dble Aw ALIGNED16;


static void alloc_arrays(int n)
{
   int m;

   if (ms>0)
      afree(Vw);

   m=n+1;
   Vw=amalloc(m*sizeof(*Vw),ALIGN);
   error((Vw==NULL),1,"alloc_arrays [langevin.c]",
         "Unable to allocate auxiliary arrays");
   cm3x3_zero(m,Vw);
   ms=m;
}


static void update_ud(int n,su3_alg_dble *frc,su3_dble *ud)
{
   int bc,m,ix,t,ifc;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            mexp_su3alg(m,frc,Vw);
            cm3x3_varmul_poly(m,Vw,ud,ud);
         }

         frc+=m;
         ud+=m;
      }
   }
}


static void update_frc(int n,double c,su3_alg_dble *frc)
{
   int bc,m,i,ix,t,ifc;
   double r;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            if ((t>0)||(ifc<2))
               r=c;
            else
               r=2.0*c;

            for (i=n;i>=2;i--)
               su3alg_mulr_add(&r,frc+i-2,frc+i);
         }

         frc+=m;
      }
   }
}


static void add_noise1(int n,double c,su3_alg_dble *frc)
{
   int bc,m,ix,t,ifc;
   double wc;

   bc=bc_type();
   m=n+1;
   wc=2.0*c;
   c=sqrt(2.0)*c;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            random_alg(1,&Aw);

            if ((t>0)||(ifc<2))
               su3alg_mulr_add(&c,&Aw,frc);
            else
               su3alg_mulr_add(&wc,&Aw,frc);
         }

         frc+=m;
      }
   }
}


static void add_noise2(int n,int nwf,double c,su3_alg_dble *eta,
                       su3_alg_dble *frc)
{
   int bc,m,ix,t,ifc;
   double wc;

   bc=bc_type();
   m=n+1;
   wc=2.0*c;
   c=sqrt(2.0)*c;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            if (nwf==1)
               random_alg(1,eta);

            if ((t>0)||(ifc<2))
               su3alg_mulr_add(&c,eta,frc);
            else
               su3alg_mulr_add(&wc,eta,frc);
         }
         else if (nwf==1)
            set_alg2zero(1,eta);

         frc+=m;
         eta+=1;
      }
   }
}


static void lang_euler(int n,su3_dble *ud)
{
   int m;
   double lambda0,eps;
   su3_alg_dble **wad,**wcd,*frc;
   lang_parms_t lang;

   m=n+1;
   lang=lang_parms();
   eps=lang.eps;
   lambda0=lang.lambda0;

   if (m>ms)
      alloc_arrays(n);

   wad=reserve_wad(m);
   frc=wad[0];

   set_alg2zero(4*VOLUME*m,frc);
   force0(n,ud,eps,frc);

   add_noise1(n,sqrt(eps),frc);
   update_ud(n,frc,ud);

   release_wad();

   wcd=reserve_wcd(m);
   add_gdfrc(n,0.0,lambda0,ud,wcd[0]);
   apply_gdfld1(n,eps,wcd[0],ud);
   release_wcd();
}


static void lang_rk2(int n,su3_dble *ud)
{
   int m;
   double k1,k2,k3,eps,lambda0;
   su3_dble **wvd;
   su3_alg_dble **wad,**wcd,*frc,*eta;
   lang_parms_t lang;

   m=n+1;
   lang=lang_parms();
   eps=lang.eps;
   lambda0=lang.lambda0;

   if (m>ms)
      alloc_arrays(n);

   wvd=reserve_wud(m);
   wad=reserve_wad(m+1);
   frc=wad[0];
   eta=wad[m];

   k1=(1.0-0.5*sqrt(2.0));
   k2=k1*k1;
   k3=1.5*k1-0.25;

   set_alg2zero(4*VOLUME*m,frc);
   force0(n,ud,k2*eps,frc);

   add_noise2(n,1,k1*sqrt(eps),eta,frc);
   cm3x3_assign(4*VOLUME*m,ud,wvd[0]);
   update_ud(n,frc,wvd[0]);

   set_alg2zero(4*VOLUME*m,frc);
   force0(n,wvd[0],eps,frc);

   update_frc(n,k3*eps,frc);
   add_noise2(n,0,sqrt(eps),eta,frc);
   update_ud(n,frc,ud);

   release_wud();
   release_wad();

   wcd=reserve_wcd(m);
   add_gdfrc(n,0.0,lambda0,ud,wcd[0]);
   apply_gdfld1(n,eps,wcd[0],ud);
   release_wcd();
}


void run_langevin(int n,su3_dble *ud)
{
   int rule;
   lang_parms_t lang;

   lang=lang_parms();
   rule=lang.rule;

   if (rule==1)
      lang_euler(n,ud);
   else if (rule==2)
      lang_rk2(n,ud);
   else
      error(1,1,"run_langevin [langevin.c]","Unknown integrator");
}
