
/*******************************************************************************
*
* File gdamp.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Gauge damping functions.
*
*   void apply_gdfld1(int n,double c,su3_alg_dble *om,su3_dble *ud)
*     Applies the gauge transformation exp(c*g0*om) to the link field ud.
*
*   void apply_gdfld2(int n,double c,su3_alg_dble *om,su3_alg_dble *mom,
*                     su3_dble *ud)
*     Applies the gauge transformation exp(c*g0*om) to the momentum field
*     mom and the link field ud.
*
*   void add_gdfrc(int n,double c1,double c2,su3_dble *ud,su3_alg_dble *om)
*     Multiplies the gauge-damping force by c2 and adds the force to the
*     field c1*om.
*
* In the program add_gdfrc(), the gauge-damping parameter lambda_0 is omitted
* and is, instead, to be included in the factor c2.
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* arrays mom and om similarly contain the n+1 coefficient fields in the
* expansions
*
*   pi=pi[0]+g0*pi[1]+..+g0^n*pi[n],
*
*   om=om[0]+g0*om[1]+..+g0^n*om[n],
*
* of the momentum and gauge-transformation fields. The programs apply_gdfld*()
* transform the field variables on the active links only.
*
* The workspaces required are
*
*   apply_gdfld*():     n+1 su3_dble point fields,
*
*   add_gdfrc():        n+1 su3_alg_dble link fields
*
* (see utils/wspace.c).
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define GDAMP_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "aflds.h"
#include "cflds.h"
#include "uflds.h"
#include "linalg.h"
#include "mdint.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ms=0;
static int nsm[NMAX],ism[NMAX];
static double *rt;
static su3_alg_dble *Ct0,*Aw;
static su3_dble *Xw,*Yw;


static void alloc_arrays(int n)
{
   int m,i;
   double r;

   if (ms==0)
   {
      rt=malloc(N0*sizeof(*rt));
      error(rt==NULL,1,"alloc_arrays [gdamp.c]",
            "Unable to allocate auxiliary arrays");

      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }

      r=1.0/((double)(N0*N0)*(double)(N1*N2*N3));

      for (i=0;i<N0;i++)
         rt[i]=(1.0-(double)(i)/(double)(N0))*r;
   }

   if (ms>0)
   {
      afree(Ct0);
      afree(Xw);
   }

   m=n+1;
   Ct0=amalloc(2*m*sizeof(*Ct0),ALIGN);
   Xw=amalloc(2*m*sizeof(*Xw),ALIGN);
   error((Ct0==NULL)||(Xw==NULL),2,"alloc_arrays [gdamp.c]",
         "Unable to allocate auxiliary arrays");
   Aw=Ct0+m;
   Yw=Xw+m;

   set_alg2zero(2*m,Ct0);
   cm3x3_zero(2*m,Xw);
   ms=m;
}


static int set_ism(int n)
{
   int m,i;

   m=n+1;

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(8*m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [gdamp.c]","Requested too many hsum instances");

   return 0;
}


static void set_gd(int n,double c,su3_alg_dble *om,su3_dble *gd)
{
   int m,ix;

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   cpcd_int_bnd(n,om);

   for (ix=0;ix<NSPIN;ix++)
   {
      mulr_alg(m,c,om,Aw);
      pexp_su3alg(m,Aw,gd);

      om+=m;
      gd+=m;
   }
}


void apply_gdfld1(int n,double c,su3_alg_dble *om,su3_dble *ud)
{
   int bc,m,ix,iy,t,ifc;
   su3_dble **wgd,*gd;

   bc=bc_type();
   m=n+1;
   wgd=reserve_wgd(m);
   gd=wgd[0];
   set_gd(n,c,om,gd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];

               if ((t==0)&&(ifc==1))
                  cm3x3_varmul_poly(m,gd+iy*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+ix*m,Yw);
                  cm3x3_varmul_poly(m,ud,Yw,Yw);
                  cm3x3_varmul_poly(m,gd+iy*m,Yw,ud);
               }
            }
            else
            {
               iy=iup[ix][ifc/2];

               if ((t==(N0-1))&&(ifc==0))
                  cm3x3_varmul_poly(m,gd+ix*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+iy*m,Yw);
                  cm3x3_varmul_poly(m,ud,Yw,Yw);
                  cm3x3_varmul_poly(m,gd+ix*m,Yw,ud);
               }
            }
         }

         ud+=m;
      }
   }

   release_wgd();
}


void apply_gdfld2(int n,double c,su3_alg_dble *om,su3_alg_dble *mom,
                  su3_dble *ud)
{
   int bc,m,ix,iy,t,ifc,i;
   su3_dble **wgd,*gd;

   bc=bc_type();
   m=n+1;
   wgd=reserve_wgd(m);
   gd=wgd[0];
   set_gd(n,c,om,gd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for (i=0;i<m;i++)
               su3alg_to_su3(mom+i,Xw+i);

            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];

               cm3x3_dagger(m,gd+iy*m,Yw);
               cm3x3_varmul_poly(m,Xw,Yw,Yw);
               cm3x3_varmul_poly(m,gd+iy*m,Yw,Xw);

               if ((t==0)&&(ifc==1))
                  cm3x3_varmul_poly(m,gd+iy*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+ix*m,Yw);
                  cm3x3_varmul_poly(m,ud,Yw,Yw);
                  cm3x3_varmul_poly(m,gd+iy*m,Yw,ud);
               }
            }
            else
            {
               iy=iup[ix][ifc/2];

               cm3x3_dagger(m,gd+ix*m,Yw);
               cm3x3_varmul_poly(m,Xw,Yw,Yw);
               cm3x3_varmul_poly(m,gd+ix*m,Yw,Xw);

               if ((t==(N0-1))&&(ifc==0))
                  cm3x3_varmul_poly(m,gd+ix*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+iy*m,Yw);
                  cm3x3_varmul_poly(m,ud,Yw,Yw);
                  cm3x3_varmul_poly(m,gd+ix*m,Yw,ud);
               }
            }

            for (i=0;i<m;i++)
               su3_to_su3alg(Xw+i,mom+i);
         }

         mom+=m;
         ud+=m;
      }
   }

   release_wgd();
}


static void set_ad(int n,su3_dble *ud,su3_alg_dble *ad)
{
   int bc,m,is,ix,ifc,i;

   bc=bc_type();
   m=n+1;

   if (bc==1)
   {
      is=set_ism(n);
      reset_hsum(is);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            for (i=0;i<m;i++)
               su3_to_su3alg(ud+i,ad+i);

            if (ifc<2)
               add_to_hsum(is,(double*)(ad));

            ud+=m;
            ad+=m;
         }
      }

      global_hsum(is,(double*)(Ct0));
   }
   else
   {
      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            for (i=0;i<m;i++)
               su3_to_su3alg(ud+i,ad+i);

            ud+=m;
            ad+=m;
         }
      }
   }
}


static void set_om(int n,double c1,double c2,su3_alg_dble *ad,su3_alg_dble *om)
{
   int bc,m,ix,iy,t,mu;

   bc=bc_type();
   m=n+1;
   c2=-c2;
   set_alg2zero((BNDRY/2)*m,om+VOLUME*m);

   if (bc==1)
   {
      for (ix=0;ix<(VOLUME/2);ix++)
      {
         t=global_time(ix);

         mulr_alg(m,c1,om+ix*m,om+ix*m);
         mulr_add_alg(m,c2*rt[t],Ct0,om+ix*m);
      }

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         mulr_alg(m,c1,om+ix*m,om+ix*m);
         mulr_add_alg(m,c2*rt[t],Ct0,om+ix*m);

         if (t!=0)
            mulr_add_alg(m,c2,ad,om+ix*m);

         if (t!=(N0-1))
         {
            iy=iup[ix][0];
            mulr_add_alg(m,-c2,ad,om+iy*m);
         }

         ad+=m;

         if (t!=0)
            mulr_add_alg(m,-c2,ad,om+ix*m);

         if (t!=1)
         {
            iy=idn[ix][0];
            mulr_add_alg(m,c2,ad,om+iy*m);
         }

         ad+=m;

         if (t!=0)
         {
            for (mu=1;mu<4;mu++)
            {
               mulr_add_alg(m,c2,ad,om+ix*m);
               iy=iup[ix][mu];
               mulr_add_alg(m,-c2,ad,om+iy*m);
               ad+=m;

               mulr_add_alg(m,-c2,ad,om+ix*m);
               iy=idn[ix][mu];
               mulr_add_alg(m,c2,ad,om+iy*m);
               ad+=m;
            }
         }
         else
            ad+=6*m;
      }
   }
   else
   {
      mulr_alg((VOLUME/2)*m,c1,om,om);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         mulr_alg(m,c1,om+ix*m,om+ix*m);

         if (t!=0)
            mulr_add_alg(m,c2,ad,om+ix*m);
         else
            mulr_add_alg(m,2.0*c2,ad,om+ix*m);

         if (t!=(N0-1))
         {
            iy=iup[ix][0];
            mulr_add_alg(m,-c2,ad,om+iy*m);
         }

         ad+=m;

         if (t!=0)
            mulr_add_alg(m,-c2,ad,om+ix*m);

         iy=idn[ix][0];

         if (t!=1)
            mulr_add_alg(m,c2,ad,om+iy*m);
         else
            mulr_add_alg(m,2.0*c2,ad,om+iy*m);

         ad+=m;

         for (mu=1;mu<4;mu++)
         {
            mulr_add_alg(m,c2,ad,om+ix*m);
            iy=iup[ix][mu];
            mulr_add_alg(m,-c2,ad,om+iy*m);
            ad+=m;

            mulr_add_alg(m,-c2,ad,om+ix*m);
            iy=idn[ix][mu];
            mulr_add_alg(m,c2,ad,om+iy*m);
            ad+=m;
         }
      }
   }

   cpcd_ext_bnd(n,om);
}


void add_gdfrc(int n,double c1,double c2,su3_dble *ud,su3_alg_dble *om)
{
   int m;
   su3_alg_dble **wad,*ad;

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   wad=reserve_wad(m);
   ad=wad[0];
   set_ad(n,ud,ad);
   set_om(n,c1,c2,ad,om);
   release_wad();
}
