
/*******************************************************************************
*
* File mdint.c
*
* Copyright (C) 2015-2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Integration of the molecular-dynamics equations.
*
*   void start_smd(int n,su3_alg_dble *om,su3_alg_dble *mom,su3_dble *ud)
*     Sets the momentum field mom and the gauge field ud to random values
*     according to the free-field equilibrium distribution (see random_mom()
*     [mflds.c] and random_ud() [uflds.c]). The gauge-damping field om is
*     set to zero (see the notes).
*
*   void run_mdint(int n,su3_alg_dble *om,su3_alg_dble *mom,su3_dble *ud,
*                  int idH,double *dH)
*     Integrates the molecular-dynamics equations using the current
*     integrator (see the notes). The changes dH[k], k=0,..,n, of the
*     the k'th order Hamilton function are computed if (and only if)
*     idH=1.
*
*   double renormalize_ud(int n,su3_dble *ud)
*     Renormalizes the field ud to SU(3) and returns the maximal deviation
*     of the old from the corrected matrix elements of the field (see the
*     notes).
*
* The integrator used is the one defined by the array of elementary operations
* returned by mdsteps() (see mdsteps.c). It is assumed that the fields and the
* integrator have been properly initialized.
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* arrays om and mom similarly contain the n+1 coefficient fields in the
* expansion
*
*   pi=pi[0]+g0*pi[1]+..+g0^n*pi[n]
*
*   om=om[0]+g0*om[1]+..+g0^n*om[n]
*
* of the momentum and the gauge-damping fields.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define MDINT_C

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "uflds.h"
#include "forces.h"
#include "mdint.h"
#include "global.h"

static int ms=0;
static double *Sm0,*Sm1,*Sg0,*Sg1;
static su3_alg_dble *Aw;
static su3_dble *Uw;


void start_smd(int n,su3_alg_dble *om,su3_alg_dble *mom,su3_dble *ud)
{
   set_alg2zero((n+1)*NSPIN,om);
   random_mom(n,mom);
   random_ud(n,ud);
}


static void alloc_arrays(int n)
{
   int m;

   if (ms>0)
   {
      free(Sm0);
      afree(Aw);
      afree(Uw);
   }

   m=n+1;
   Sm0=malloc(4*m*sizeof(*Sm0));
   Aw=amalloc(m*sizeof(*Aw),4);
   Uw=amalloc(m*sizeof(*Uw),4);

   error((Sm0==NULL)||(Aw==NULL)||(Uw==NULL),1,"alloc_arrays [mdint.c]",
         "Unable to allocate auxiliary arrays");

   Sm1=Sm0+m;
   Sg0=Sm1+m;
   Sg1=Sg0+m;
   ms=m;
}


static void update_ud(int n,double eps,su3_alg_dble *mom,su3_dble *ud)
{
   int bc,m,ix,t,ifc;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            mulr_alg(m,eps,mom,Aw);
            pexp_su3alg(m,Aw,Uw);
            cm3x3_varmul_poly(m,Uw,ud,ud);
         }

         mom+=m;
         ud+=m;
      }
   }
}


void run_mdint(int n,su3_alg_dble *om,su3_alg_dble *mom,su3_dble *ud,
               int idH,double *dH)
{
   int m,nop,ifr,k;
   double co,si;
   double eps,gamma,lambda0;
   top_t top;
   smd_parms_t smd;
   mdstep_t *s,*sm;

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   smd=smd_parms();
   gamma=smd.gamma;
   lambda0=smd.lambda0;

   s=mdsteps(&nop);
   sm=s+nop;

   for (;s<sm;s++)
   {
      top=(*s).top;
      eps=(*s).eps;

      if (top==TP)
      {
         ifr=(*s).ifr;

         if (ifr==0)
            force0(n,ud,-eps,mom);
         else
            error(1,1,"run_mdint [mdint.c]","Unexpected force number");
      }
      else if (top==TU)
         update_ud(n,eps,mom,ud);
      else if (top==TO)
         add_gdfrc(n,1.0,eps*lambda0,ud,om);
      else if (top==TG)
         apply_gdfld2(n,eps,om,mom,ud);
      else if (top==TR)
      {
         co=exp(-eps*gamma);
         si=sqrt(2.0*co*sinh(eps*gamma));
         rotate_mom(n,co,si,mom);
         mulr_alg(VOLUME*m,co,om,om);

         if (idH==1)
         {
            mom_action(0,n,mom,Sm0);
            action0(0,n,ud,Sg0);
         }
      }
      else
         error(1,1,"run_mdint [mdint.c]","Unexpected operation type");
   }

   if (idH==1)
   {
      mom_action(0,n,mom,Sm1);
      action0(0,n,ud,Sg1);

      for (k=0;k<m;k++)
         dH[k]=(Sm1[k]-Sm0[k])+(Sg1[k]-Sg0[k]);

      if (NPROC>1)
      {
         for (k=0;k<m;k++)
            Sm0[k]=dH[k];

         MPI_Reduce(Sm0,dH,m,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(dH,m,MPI_DOUBLE,0,MPI_COMM_WORLD);
      }
   }
}


double renormalize_ud(int n,su3_dble *ud)
{
   int bc,m,ix,t,ifc;
   double d,dev;
   complex_dble *z,*zm;

   bc=bc_type();
   m=n+1;
   dev=0.0;

   if (m>ms)
      alloc_arrays(n);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            cm3x3_inv_poly(m,ud,Uw);
            cm3x3_ren2su3_poly(m,ud);
            cm3x3_varmul_poly(m,ud,Uw,Uw);
            z=(complex_dble*)(Uw);
            zm=z+9*m;

            for (;z<zm;z++)
            {
               d=(*z).re*(*z).re+(*z).im*(*z).im;
               if (d>dev)
                  dev=d;
            }
         }

         ud+=m;
      }
   }

   if (NPROC>1)
   {
      d=dev;
      MPI_Reduce(&d,&dev,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&dev,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   return sqrt(dev);
}
