
/*******************************************************************************
*
* File mdsteps.c
*
* Copyright (C) 2011, 2012, 2015, 2017 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Molecular-dynamics integrator.
*
*   void set_mdsteps(void)
*     Constructs the integrator from the data available in the parameter
*     data base (see the notes). The integrator is stored internally in
*     the form of an array of elementary operations (force computations
*     and gauge-field update steps).
*
*   mdstep_t *mdsteps(int *nop)
*     Returns the array of elementary operations that describes the current
*     integrator. The total number of operations is assigned to nop.
*
*   void print_mdsteps(int ipr)
*     Prints some information on the current integrator to stdout on MPI
*     process 0 depending on the 2 low bits of the print flat ipr:
*
*      If (ipr&0x1): List of elementary operations.
*
*      If (ipr&0x2): Integration time check.
*
*     The full information is thus printed if ipr=0x3.
*
* The structure of the MD integrator is explained in the file README.mdint
* in this directory. It is assumed that the parameters of the integrator
* have been entered to the parameter data base [smd_parms.c, mdint_parms.c].
*
* An elementary update step is described by a structure of type mdstep_t
* with the following elements:
*
*  top    Operation type (TP,TU,TO,TG or TR).
*
*  ifr    Number of the force to be added to the momentum field if top=TP.
*         Set to 0 if top!=TP.
*
*  eps    Integration step size at the current point of the integration.
*
* See README.mdint for further explanations.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "mdint.h"

static int nsmx,nmds=0;
static const mdstep_t md0={TOPS,0,0.0};
static mdstep_t *mds=NULL,*mdw[3];
static const char *tops[(int)(TOPS)]={"TP","TU","TO","TG","TR"};


static void set_nsmx(int nlv)
{
   int ntu,ntp,nfr,nstep;
   int mtu,ilv;
   mdint_parms_t mdp;

   nsmx=5;
   mtu=1;

   for (ilv=nlv-1;ilv>=0;ilv--)
   {
      mdp=mdint_parms(ilv);
      nstep=mdp.nstep;
      nfr=mdp.nfr;

      if (mdp.integrator==LPFR)
      {
         ntu=nstep;
         ntp=2*nstep;
      }
      else if (mdp.integrator==OMF2)
      {
         ntu=2*nstep;
         ntp=3*nstep;
      }
      else if (mdp.integrator==OMF4)
      {
         ntu=5*nstep;
         ntp=6*nstep;
      }
      else if (mdp.integrator==OMF6)
      {
         ntu=7*nstep;
         ntp=8*nstep;
      }
      else
      {
         error_loc(1,1,"set_nsmx [mdsteps.c]","Unknown integrator");
         ntu=0;
         ntp=0;
      }

      nsmx+=mtu*(ntu+ntp*nfr-1);
      mtu*=ntu;
   }
}


static void alloc_mds(void)
{
   int k;

   if (mds!=NULL)
      free(mds);

   mds=malloc(4*nsmx*sizeof(*mds));
   error(mds==NULL,1,"alloc_mds [mdsteps.c]",
         "Unable to allocate mdsteps array");

   for (k=0;k<3;k++)
      mdw[k]=mds+(k+1)*nsmx;

   for (k=0;k<(4*nsmx);k++)
      mds[k]=md0;
}


static void copy_steps(int n,double c,mdstep_t *s,mdstep_t *r)
{
   int i;

   for (i=0;i<n;i++)
   {
      r[i].top=s[i].top;
      r[i].ifr=s[i].ifr;
      r[i].eps=c*s[i].eps;
   }
}


static void expand_level(int ilv,double tau,mdstep_t *s,mdstep_t *ws)
{
   int nstep,ntu,nfr,*ifr;
   int n,i;
   double a1,a2,a3,a4,b1,b2,b3,b4;
   double r0,r1,r2,r3,r4,eps;
   mdstep_t *sP,*sU,*r;
   mdint_parms_t mdp;

   mdp=mdint_parms(ilv);
   nstep=mdp.nstep;
   nfr=mdp.nfr;
   ifr=mdp.ifr;

   a1=0.2465881872786138;
   b1=0.08333333333333333;
   a2=0.6047073875057809;
   b2=0.3977675859548440;
   a3=-0.4009869039788007;
   b3=-0.03933369314462574;
   a4=0.09938265838881204;
   b4=0.05823277385644840;

   r0=mdp.lambda;
   r1=0.08398315262876693;
   r2=0.2539785108410595;
   r3=0.6822365335719091;
   r4=-0.03230286765269967;
   eps=tau/(double)(nstep);

   sP=s;

   for (i=0;i<nfr;i++)
   {
      sP[i]=md0;
      sP[i].top=TP;
      sP[i].ifr=ifr[i];
      sP[i].eps=eps;
   }

   ntu=1;
   sU=sP+nfr;
   sU[0]=md0;
   sU[0].top=TU;
   sU[0].eps=eps;

   n=0;
   r=ws;

   if (mdp.integrator==LPFR)
   {
      copy_steps(nfr,0.5,sP,r);
      r+=nfr;
      copy_steps(ntu,1.0,sU,r);
      r+=ntu;
      copy_steps(nfr,0.5,sP,r);
      r+=nfr;
      n+=2*nfr+ntu;
   }
   else if (mdp.integrator==OMF2)
   {
      copy_steps(nfr,r0,sP,r);
      r+=nfr;
      copy_steps(ntu,0.5,sU,r);
      r+=ntu;
      copy_steps(nfr,1.0-2.0*r0,sP,r);
      r+=nfr;
      copy_steps(ntu,0.5,sU,r);
      r+=ntu;
      copy_steps(nfr,r0,sP,r);
      r+=nfr;
      n+=3*nfr+2*ntu;
   }
   else if (mdp.integrator==OMF4)
   {
      copy_steps(nfr,r1,sP,r);
      r+=nfr;
      copy_steps(ntu,r2,sU,r);
      r+=ntu;
      copy_steps(nfr,r3,sP,r);
      r+=nfr;
      copy_steps(ntu,r4,sU,r);
      r+=ntu;
      copy_steps(nfr,0.5-r1-r3,sP,r);
      r+=nfr;
      copy_steps(ntu,1.0-2.0*(r2+r4),sU,r);
      r+=ntu;
      copy_steps(nfr,0.5-r1-r3,sP,r);
      r+=nfr;
      copy_steps(ntu,r4,sU,r);
      r+=ntu;
      copy_steps(nfr,r3,sP,r);
      r+=nfr;
      copy_steps(ntu,r2,sU,r);
      r+=ntu;
      copy_steps(nfr,r1,sP,r);
      r+=nfr;
      n+=6*nfr+5*ntu;
   }
   else if (mdp.integrator==OMF6)
   {
      copy_steps(nfr,b1,sP,r);
      r+=nfr;
      copy_steps(ntu,a1,sU,r);
      r+=ntu;
      copy_steps(nfr,b2,sP,r);
      r+=nfr;
      copy_steps(ntu,a2,sU,r);
      r+=ntu;
      copy_steps(nfr,b3,sP,r);
      r+=nfr;
      copy_steps(ntu,a3,sU,r);
      r+=ntu;
      copy_steps(nfr,b4,sP,r);
      r+=nfr;
      copy_steps(ntu,a4,sU,r);
      r+=ntu;
      copy_steps(nfr,b4,sP,r);
      r+=nfr;
      copy_steps(ntu,a3,sU,r);
      r+=ntu;
      copy_steps(nfr,b3,sP,r);
      r+=nfr;
      copy_steps(ntu,a2,sU,r);
      r+=ntu;
      copy_steps(nfr,b2,sP,r);
      r+=nfr;
      copy_steps(ntu,a1,sU,r);
      r+=ntu;
      copy_steps(nfr,b1,sP,r);
      r+=nfr;
      n+=8*nfr+7*ntu;
   }

   r=s;

   for (i=0;i<nstep;i++)
   {
      copy_steps(n,1.0,ws,r);
      r+=n;
   }

   (*r)=md0;
}


static int nall_steps(mdstep_t *s)
{
   int n;

   for (n=0;s[n].top!=TOPS;n++);

   return n;
}


static void combine_forces(mdstep_t *s)
{
   int n,i,j;

   n=0;

   for (i=0;s[i].top!=TOPS;i++)
   {
      if (s[i].top==TP)
      {
         for (j=i+1;s[j].top==TP;j++)
         {
            if (s[i].ifr==s[j].ifr)
            {
               s[j].eps+=s[i].eps;
               s[i].eps=0.0;
               break;
            }
         }
      }

      if (s[i].eps!=0.0)
      {
         s[n]=s[i];
         n+=1;
      }
   }

   s[n]=md0;
}


static void insert_level(mdstep_t *s1,mdstep_t *s2,mdstep_t *r)
{
   int nall,i;
   double eps;

   nall=nall_steps(s1);

   for (i=0;s2[i].top!=TOPS;i++)
   {
      if (s2[i].top==TU)
      {
         eps=s2[i].eps;
         copy_steps(nall,eps,s1,r);
         r+=nall;
      }
      else
      {
         (*r)=s2[i];
         r+=1;
      }
   }

   (*r)=md0;
}


void set_mdsteps(void)
{
   int nlv,ilv,n;
   double eps;
   smd_parms_t smd;

   smd=smd_parms();
   nlv=smd.nlv;
   eps=smd.eps;

   set_nsmx(nlv);
   alloc_mds();

   mds[0].top=TR;
   mds[0].ifr=0;
   mds[0].eps=eps;
   expand_level(nlv-1,eps,mds+1,mdw[0]);
   combine_forces(mds);

   for (ilv=(nlv-2);ilv>=0;ilv--)
   {
      n=nall_steps(mds);
      copy_steps(n+1,1.0,mds,mdw[0]);
      expand_level(ilv,1.0,mdw[1],mdw[2]);
      insert_level(mdw[1],mdw[0],mds);
      combine_forces(mds);
   }

   n=nall_steps(mds);

   mds[n]=md0;
   mds[n].top=TO;
   mds[n].eps=eps;
   n+=1;
   mds[n]=md0;
   mds[n].top=TG;
   mds[n].eps=eps;
   n+=1;
   mds[n]=md0;

   nmds=n;
}


mdstep_t *mdsteps(int *nop)
{
   (*nop)=nmds;

   return mds;
}


static void print_ops(void)
{
   int i;

   printf("List of elementary operations:\n");

   for (i=0;i<nmds;i++)
   {
      if (mds[i].top==TP)
         printf("TP: eps = % .2e, force no %2d\n",mds[i].eps,mds[i].ifr);
      else if (mds[i].top==TU)
         printf("TU: eps = % .2e\n",mds[i].eps);
      else if (mds[i].top==TR)
         printf("TR: eps = % .2e\n",mds[i].eps);
      else if (mds[i].top==TO)
         printf("TO: eps = % .2e\n",mds[i].eps);
      else if (mds[i].top==TG)
         printf("TG: eps = % .2e\n",mds[i].eps);
      else
         error_root(1,1,"print_ops [mdsteps.c]","Unkown operation");
   }

   printf("\n");
}


static void print_times(double eps)
{
   int i,j,k;
   int nfr,it;
   double seps;
   top_t top;

   printf("Total integration times:\n");

   for (i=0;i<(int)(TOPS);i++)
   {
      top=(top_t)(i);

      if (top==TP)
      {
         nfr=0;

         for (j=0;j<nmds;j++)
         {
            if ((mds[j].top==TP)&&(mds[j].ifr>nfr))
               nfr=mds[j].ifr;
         }

         for (j=0;j<=nfr;j++)
         {
            it=0;
            seps=0.0;

            for (k=0;k<nmds;k++)
            {
               if ((mds[k].top==TP)&&(mds[k].ifr==j))
               {
                  it=1;
                  seps+=mds[k].eps;
               }
            }

            if (it)
               printf("Force no %2d: sum(dt)/eps = %.3e\n",j,seps/eps);
         }
      }
      else
      {
         it=0;
         seps=0.0;

         for (j=0;j<nmds;j++)
         {
            if (mds[j].top==top)
            {
               it=1;
               seps+=mds[j].eps;
            }
         }

         if (it)
            printf("%s:          sum(dt)/eps = %.3e\n",tops[i],seps/eps);
      }
   }

   printf("\n");
}


void print_mdsteps(int ipr)
{
   int my_rank;
   smd_parms_t smd;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   smd=smd_parms();

   if (my_rank==0)
   {
      if (ipr&0x1)
         print_ops();

      if (ipr&0x2)
         print_times(smd.eps);
   }
}
