
/*******************************************************************************
*
* File smd_parms.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Parameters of the SMD algorithm.
*
*   smd_parms_t set_smd_parms(int nact,int *iact,int npf,int nmu,double *mu,
*                             int nlv,double eps,double gamma,double lambda0)
*     Sets the basic parameters of the SMD algorithm. The parameters are
*
*       nact        Number of terms in the total action.
*
*       iact        Indices iact[i] of the action terms (i=0,..,nact-1).
*
*       npf         Number of pseudo-fermion fields on which the action
*                   depends.
*
*       nmu         Number of twisted mass parameters on which the
*                   pseudo-fermion actions and forces depend.
*
*       mu          Twisted masses mu[i] (i=0,..,nmu-1).
*
*       nlv         Number of levels of the molecular-dynamics integrator.
*
*       eps         Simulation step size.
*
*       gamma       Momentum refreshment parameter.
*
*       lambda0     Gauge damping parameter.
*
*     The total action must include the gauge action, but pseudo-fermion
*     actions are optional and the momentum action is treated separately.
*     The program returns a structure containing the parameters listed
*     above.
*
*   smd_parms_t smd_parms(void)
*     Returns a structure containing the current values of the parameters
*     listed above.
*
*   void print_smd_parms(void)
*     Prints the SMD parameters to stdout on MPI process 0.
*
*   void write_smd_parms(FILE *fdat)
*     Writes the SMD parameters to the file fdat on MPI process 0.
*
*   void check_smd_parms(FILE *fdat)
*     Compares the SMD parameters with the values stored on the file fdat
*     on MPI process 0, assuming the latter were written to the file by
*     the program write_smd_parms().
*
* The parameters eps, gamma and lambda0 must all be positive. To ensure the
* consistency of the data base, the parameters must be set simultaneously on
* all processes. The type smd_parms_t is defined in the file pmdb.h.
*
*******************************************************************************/

#define SMD_PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

static smd_parms_t smd={0,0,0,0,NULL,0.0,0.0,0.0,NULL};


smd_parms_t set_smd_parms(int nact,int *iact,int npf,int nmu,double *mu,
                          int nlv,double eps,double gamma,double lambda0)
{
   int iprms[4],i,ie;
   double dprms[3];

   if (NPROC>1)
   {
      iprms[0]=nact;
      iprms[1]=npf;
      iprms[2]=nmu;
      iprms[3]=nlv;

      dprms[0]=eps;
      dprms[1]=gamma;
      dprms[2]=lambda0;

      MPI_Bcast(iprms,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,3,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=nact)||(iprms[1]!=npf)||(iprms[2]!=nmu)||
            (iprms[3]!=nlv)||(dprms[0]!=eps)||(dprms[1]!=gamma)||
            (dprms[2]!=lambda0),1,"set_smd_parms [smd_parms.c]",
            "Parameters are not global");

      ie=0;

      for (i=0;i<nact;i++)
      {
         iprms[0]=iact[i];
         MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
         ie|=(iprms[0]!=iact[i]);
      }

      for (i=0;i<nmu;i++)
      {
         dprms[0]=mu[i];
         MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
         ie|=(dprms[0]!=mu[i]);
      }

      error(ie!=0,2,"set_smd_parms [smd_parms.c]",
            "Parameters are not global");
   }

   error_root((npf<0)||(nlv<1),1,"set_smd_parms [smd_parms.c]",
              "Improper number of pseudo-fermion fields or integrator levels");
   error_root((smd.nlv>0)&&(npf!=smd.npf),1,"set_smd_parms [smd_parms.c]",
              "Number of pseudo-fermion fields may be set only once");
   error_root((eps<=0.0)||(gamma<=0.0)||(lambda0<=0.0),1,
              "set_smd_parms [smd_parms.c]",
              "Parameters eps, gamma and lambda0 must be positive");

   if (nact!=smd.nact)
   {
      if (smd.iact!=NULL)
      {
         free(smd.iact);
         smd.iact=NULL;
      }

      if (nact>0)
      {
         smd.iact=malloc(nact*sizeof(int));
         error(smd.iact==NULL,1,"set_smd_parms [smd_parms.c]",
               "Unable to allocate parameter array");
      }
   }

   if (nmu!=smd.nmu)
   {
      if (smd.mu!=NULL)
      {
         free(smd.mu);
         smd.mu=NULL;
      }

      if (nmu>0)
      {
         smd.mu=malloc(nmu*sizeof(double));
         error(smd.mu==NULL,2,"set_smd_parms [smd_parms.c]",
               "Unable to allocate parameter array");
      }
   }

   smd.nact=nact;
   smd.npf=npf;
   smd.nmu=nmu;
   smd.nlv=nlv;
   smd.eps=eps;
   smd.gamma=gamma;
   smd.lambda0=lambda0;

   for (i=0;i<nact;i++)
      smd.iact[i]=iact[i];

   for (i=0;i<nmu;i++)
      smd.mu[i]=mu[i];

   return smd;
}


smd_parms_t smd_parms(void)
{
   return smd;
}


void print_smd_parms(void)
{
   int my_rank,n,i;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      printf("SMD parameters:\n");
      printf("actions =");
      for (i=0;i<smd.nact;i++)
         printf(" %d",smd.iact[i]);
      printf("\n");
      printf("npf = %d\n",smd.npf);
      if (smd.nmu>0)
      {
         printf("mu =");
         for (i=0;i<smd.nmu;i++)
         {
            n=fdigits(smd.mu[i]);
            printf(" %.*f",IMAX(n,1),smd.mu[i]);
         }
         printf("\n");
      }
      printf("nlv = %d\n",smd.nlv);
      n=fdigits(smd.eps);
      printf("eps = %.*f\n",IMAX(n,1),smd.eps);
      n=fdigits(smd.gamma);
      printf("gamma = %.*f\n",IMAX(n,1),smd.gamma);
      n=fdigits(smd.lambda0);
      printf("lambda0 = %.*f\n\n",IMAX(n,1),smd.lambda0);
   }
}


void write_smd_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,iw;
   stdint_t istd[4];
   double dstd[3];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      istd[0]=(stdint_t)(smd.nact);
      istd[1]=(stdint_t)(smd.npf);
      istd[2]=(stdint_t)(smd.nmu);
      istd[3]=(stdint_t)(smd.nlv);

      dstd[0]=smd.eps;
      dstd[1]=smd.gamma;
      dstd[2]=smd.lambda0;

      if (endian==BIG_ENDIAN)
      {
         bswap_int(4,istd);
         bswap_double(3,dstd);
      }

      iw=fwrite(istd,sizeof(stdint_t),4,fdat);
      iw+=fwrite(dstd,sizeof(double),3,fdat);

      for (i=0;i<smd.nact;i++)
      {
         istd[0]=(stdint_t)(smd.iact[i]);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         iw+=fwrite(istd,sizeof(stdint_t),1,fdat);
      }

      for (i=0;i<smd.nmu;i++)
      {
         dstd[0]=smd.mu[i];

         if (endian==BIG_ENDIAN)
            bswap_double(1,dstd);

         iw+=fwrite(dstd,sizeof(double),1,fdat);
      }

      error_root(iw!=(7+smd.nact+smd.nmu),1,"write_smd_parms [smd_parms.c]",
                 "Incorrect write count");
   }
}


void check_smd_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,ie,ir;
   stdint_t istd[4];
   double dstd[3];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      ir=fread(istd,sizeof(stdint_t),4,fdat);
      ir+=fread(dstd,sizeof(double),3,fdat);
      error_root(ir!=7,1,"check_smd_parms [smd_parms.c]",
                 "Incorrect read count");

      if (endian==BIG_ENDIAN)
      {
         bswap_int(4,istd);
         bswap_double(3,dstd);
      }

      ie=0;
      ie|=(istd[0]!=(stdint_t)(smd.nact));
      ie|=(istd[1]!=(stdint_t)(smd.npf));
      ie|=(istd[2]!=(stdint_t)(smd.nmu));
      ie|=(istd[3]!=(stdint_t)(smd.nlv));

      ie|=(dstd[0]!=smd.eps);
      ie|=(dstd[1]!=smd.gamma);
      ie|=(dstd[2]!=smd.lambda0);

      for (i=0;i<smd.nact;i++)
      {
         ir+=fread(istd,sizeof(stdint_t),1,fdat);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         ie|=(istd[0]!=(stdint_t)(smd.iact[i]));
      }

      for (i=0;i<smd.nmu;i++)
      {
         ir+=fread(dstd,sizeof(double),1,fdat);

         if (endian==BIG_ENDIAN)
            bswap_double(1,dstd);

         ie|=(dstd[0]!=smd.mu[i]);
      }

      error_root(ir!=(7+smd.nact+smd.nmu),1,"check_smd_parms [smd_parms.c]",
                 "Incorrect read count");
      error_root(ie!=0,1,"check_smd_parms [smd_parms.c]",
                 "Parameters do not match");
   }
}
