
/*******************************************************************************
*
* File wflow.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Integration of the Wilson flow.
*
*   void fwd_euler(int ns,double eps,int idmp,int n,su3_dble *ud)
*     Applies ns forward Euler integration steps, with step size eps, to
*     the link field ud (see the notes).
*
*   void fwd_rk2(int ns,double eps,int idmp,int n,su3_dble *ud)
*     Applies ns forward 2nd-order Runge-Kutta integration steps, with step
*     size eps, to the link field ud (see the notes).
*
*   void fwd_rk3(int ns,double eps,int idmp,int n,su3_dble *ud)
*     Applies ns forward 3rd-order Runge-Kutta integration steps, with step
*     size eps, to the link field ud (see the notes).
*
* The implementation of the gradient flow is described in
*
*   M. Luescher, "Yang-Mills gradient flow in stochastic perturbation
*                 theory", notes (May 2015), file doc/gflow.pdf.
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* gauge damping term is included in the flow equation if (and only if) the
* parameter idmp is set 1.
*
* The Runge-Kutta integrators require a workspace of 3*(n+1) su(3) link
* fields and n+1 su(3) ghost fields (this includes the workspace needed for
* the force computation). The Euler integrator requires only 2*(n+1) su(3)
* link fields instead of 3*(n+1) fields.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define WFLOW_C

#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "su3fcts.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "cflds.h"
#include "uflds.h"
#include "linalg.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int ms=0;
static su3_alg_dble *Aw;
static su3_dble *Vw;


static void alloc_arrays(int n)
{
   int m;

   if (ms>0)
   {
      afree(Aw);
      afree(Vw);
   }

   m=n+1;
   Aw=amalloc(m*sizeof(*Aw),ALIGN);
   Vw=amalloc(m*sizeof(*Vw),ALIGN);
   error((Aw==NULL)||(Vw==NULL),1,"alloc_arrays [wflow.c]",
         "Unable to allocate auxiliary arrays");

   set_alg2zero(m,Aw);
   cm3x3_zero(m,Vw);
   ms=m;
}


static void update_ud(double eps,int n,su3_alg_dble *fd,su3_dble *ud)
{
   int bc,m,ix,t,ifc;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            mulr_alg(m,eps,fd,Aw);
            pexp_su3alg(m,Aw,Vw);
            cm3x3_varmul_poly(m,Vw,ud,ud);
         }

         fd+=m;
         ud+=m;
      }
   }
}


static void update_fro1(int n,double c,su3_alg_dble *frc,su3_alg_dble *fro)
{
   su3_alg_dble *frm;

   frm=frc+4*VOLUME*(n+1);

   for (;frc<frm;frc++)
   {
      su3alg_mulr_add(&c,frc,fro);
      fro+=1;
   }
}


static void update_fro2(int n,double c,su3_alg_dble *frc,su3_alg_dble *fro)
{
   su3_alg_dble *frm;

   frm=frc+4*VOLUME*(n+1);

   for (;frc<frm;frc++)
   {
      su3alg_mulr(&c,fro,fro);
      su3alg_add(frc,fro);
      fro+=1;
   }
}


void fwd_euler(int ns,double eps,int idmp,int n,su3_dble *ud)
{
   int m,k,iprms[3];
   double dprms[1];
   su3_alg_dble **wad,*frc;

   error_root(n<0,1,"fwd_euler [wflow.c]","Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=ns;
      iprms[1]=idmp;
      iprms[2]=n;
      dprms[0]=eps;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=ns)||(iprms[1]!=idmp)||(iprms[2]!=n)||(dprms[0]!=eps),
            1,"fwd_euler [wflow.c]","Parameters are not global");
   }

   if (ns>0)
   {
      m=n+1;

      if (m>ms)
         alloc_arrays(n);

      wad=reserve_wad(m);
      frc=wad[0];

      for (k=0;k<ns;k++)
      {
         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         update_ud(eps,n,frc,ud);
      }

      release_wad();
   }
}


void fwd_rk2(int ns,double eps,int idmp,int n,su3_dble *ud)
{
   int m,k,iprms[3];
   double dprms[1];
   su3_alg_dble **wad,*frc,*fro;

   error_root(n<0,1,"fwd_rk2 [wflow.c]","Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=ns;
      iprms[1]=idmp;
      iprms[2]=n;
      dprms[0]=eps;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=ns)||(iprms[1]!=idmp)||(iprms[2]!=n)||(dprms[0]!=eps),
            1,"fwd_rk2 [wflow.c]","Parameters are not global");
   }

   if (ns>0)
   {
      m=n+1;

      if (m>ms)
         alloc_arrays(n);

      wad=reserve_wad(2*m);
      frc=wad[0];
      fro=wad[m];

      for (k=0;k<ns;k++)
      {
         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         assign_alg2alg(4*VOLUME*m,frc,fro);
         update_ud(0.5*eps,n,frc,ud);

         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         update_fro2(n,-0.5,frc,fro);
         update_ud(eps,n,fro,ud);
      }

      release_wad();
   }
}


void fwd_rk3(int ns,double eps,int idmp,int n,su3_dble *ud)
{
   int m,k,iprms[3];
   double dprms[1];
   su3_alg_dble **wad,*frc,*fro;

   error_root(n<0,1,"fwd_rk3 [wflow.c]","Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=ns;
      iprms[1]=idmp;
      iprms[2]=n;
      dprms[0]=eps;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=ns)||(iprms[1]!=idmp)||(iprms[2]!=n)||(dprms[0]!=eps),
            1,"fwd_rk3 [wflow.c]","Parameters are not global");
   }

   if (ns>0)
   {
      m=n+1;

      if (m>ms)
         alloc_arrays(n);

      wad=reserve_wad(2*m);
      frc=wad[0];
      fro=wad[m];

      for (k=0;k<ns;k++)
      {
         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         assign_alg2alg(4*VOLUME*m,frc,fro);
         update_ud(0.25*eps,n,frc,ud);

         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         update_fro1(n,-32.0/17.0,frc,fro);
         update_ud((-17.0/36.0)*eps,n,fro,ud);

         plaq_frc(n,ud,frc);
         if (idmp==1)
            damping_frc(n,ud,frc);
         update_fro2(n,17.0/27.0,frc,fro);
         update_ud(0.75*eps,n,fro,ud);
      }

      release_wad();
   }
}
